@@ -2,10 +2,8 @@
 # this repository contains the full copyright notices and license terms.
 
 from trytond.pool import Pool
-from . import account
-from . import ir
-from . import party
-from . import routes
+
+from . import account, ir, party, routes
 
 __all__ = ['register', 'routes']
 
@@ -19,8 +17,10 @@ def register():
         account.PaymentBraintreeRefund,
         account.PaymentBraintreeAccount,
         account.PaymentBraintreeCustomer,
+        account.PaymentBraintreeCustomerIdentical,
         account.PaymentBraintreeCustomerPaymentMethodDeleteAsk,
         party.Party,
+        party.PartyReceptionDirectDebit,
         module='account_payment_braintree', type_='model')
     Pool.register(
         account.PaymentBraintreeCustomerPaymentMethodDelete,
