@@ -7,6 +7,7 @@ import uuid
 import braintree
 from braintree.exceptions import TooManyRequestsError
 from braintree.exceptions.braintree_error import BraintreeError
+from sql import Literal
 
 try:
     from braintree.exceptions import GatewayTimeoutError
@@ -18,17 +19,22 @@ from trytond.cache import Cache
 from trytond.config import config
 from trytond.i18n import gettext
 from trytond.model import (
-    ModelSQL, ModelView, DeactivableMixin, Workflow, fields)
-from trytond.pool import PoolMeta, Pool
-from trytond.pyson import Eval, Bool
+    DeactivableMixin, Index, ModelSQL, ModelView, Workflow, fields)
+from trytond.modules.account_payment.exceptions import (
+    PaymentValidationError, ProcessError)
+from trytond.modules.company.model import (
+    employee_field, reset_employee, set_employee)
+from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval
 from trytond.report import Report
+from trytond.tools import sql_pairing
 from trytond.transaction import Transaction
 from trytond.url import http_host
 from trytond.wizard import (
-    Wizard, StateAction, StateView, StateTransition, Button)
+    Button, StateAction, StateTransition, StateView, Wizard)
 
-from trytond.modules.account_payment.exceptions import (
-    ProcessError, PaymentValidationError)
+from .common import BraintreeCustomerMethodMixin
 
 logger = logging.getLogger(__name__)
 
@@ -59,8 +65,7 @@ class PaymentJournal(metaclass=PoolMeta):
         states={
             'required': Eval('process_method') == 'braintree',
             'invisible': Eval('process_method') != 'braintree',
-            },
-        depends=['currency', 'process_method'])
+            })
 
     @classmethod
     def __setup__(cls):
@@ -73,6 +78,12 @@ class PaymentJournal(metaclass=PoolMeta):
 class PaymentGroup(metaclass=PoolMeta):
     __name__ = 'account.payment.group'
 
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls._buttons['succeed']['invisible'] |= (
+            Eval('process_method') == 'braintree')
+
     def process_braintree(self):
         pool = Pool()
         Payment = pool.get('account.payment')
@@ -94,16 +105,18 @@ class PaymentGroup(metaclass=PoolMeta):
 
 
 class CheckoutMixin:
-    braintree_checkout_id = fields.Char("Braintree Checkout ID", readonly=True)
+    __slots__ = ()
+    braintree_checkout_id = fields.Char(
+        "Braintree Checkout ID", readonly=True, strip=False)
     braintree_client_token = fields.Char(
-        "Braintree Client Token", readonly=True)
+        "Braintree Client Token", readonly=True, strip=False)
     braintree_nonce = fields.Char(
-        "Braintree Nonce", readonly=True,
+        "Braintree Nonce", readonly=True, strip=False,
         states={
             'invisible': ~Eval('braintree_nonce'),
             })
     braintree_device_data = fields.Char(
-        "Braintree Device Data", readonly=True,
+        "Braintree Device Data", readonly=True, strip=False,
         states={
             'invisible': ~Eval('braintree_device_data'),
             })
@@ -147,71 +160,21 @@ class CheckoutMixin:
         self.save()
 
 
-class Payment(CheckoutMixin, metaclass=PoolMeta):
+class Payment(CheckoutMixin, BraintreeCustomerMethodMixin, metaclass=PoolMeta):
     __name__ = 'account.payment'
 
-    braintree_account = fields.Function(fields.Many2One(
-            'account.payment.braintree.account', "Braintree Account"),
-        'on_change_with_braintree_account')
-
-    braintree_customer = fields.Many2One(
-        'account.payment.braintree.customer', "Braintree Customer",
-        domain=[
-            ('party', '=', Eval('party', -1)),
-            ('braintree_account', '=', Eval('braintree_account', -1)),
-            ],
-        states={
-            'invisible': Eval('process_method') != 'braintree',
-            'readonly': (~Eval('state').in_(['draft', 'approved'])
-                | Eval('braintree_nonce')),
-            },
-        depends=['party', 'braintree_account', 'process_method',
-            'braintree_nonce', 'state'])
-    braintree_customer_method = fields.Char(
-        "Braintree Customer Method",
-        states={
-            'invisible': ((Eval('process_method') != 'braintree')
-                | Eval('braintree_nonce')
-                | ~Eval('braintree_customer')),
-            'readonly': ~Eval('state').in_(['draft', 'approved']),
-            },
-        depends=[
-            'process_method',
-            'braintree_nonce',
-            'braintree_customer',
-            'state'])
-    # Use Function field with selection
-    # to avoid querying Braintree on validation
-    braintree_customer_method_selection = fields.Function(fields.Selection(
-            'get_braintree_customer_methods', "Braintree Customer Method",
-            states={
-                'invisible': ((Eval('process_method') != 'braintree')
-                    | Eval('braintree_nonce')
-                    | ~Eval('braintree_customer')),
-                'readonly': ~Eval('state').in_(['draft', 'approved']),
-                },
-            depends=[
-                'process_method',
-                'braintree_nonce',
-                'braintree_customer',
-                'state']),
-        'get_braintree_customer_method',
-        setter='set_braintree_customer_method')
-
     braintree_transaction_id = fields.Char(
-        "Braintree Transaction ID", readonly=True,
+        "Braintree Transaction ID", readonly=True, strip=False,
         states={
             'invisible': ((Eval('process_method') != 'braintree')
                 | ~Eval('braintree_transaction_id')),
-            },
-        depends=['process_method'])
+            })
     braintree_settle_payment = fields.Boolean(
         "Braintree Settle Payment",
         states={
             'invisible': Eval('process_method') != 'braintree',
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['process_method', 'state'])
+            })
     braintree_payment_settled = fields.Boolean(
         "Braintree Payment Settled", readonly=True)
     braintree_settlement_needed = fields.Function(
@@ -223,8 +186,7 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
         states={
             'invisible': ((Eval('process_method') != 'braintree')
                 | ~Eval('braintree_transaction_id')),
-            },
-        depends=['process_method', 'braintree_transaction_id'])
+            })
 
     braintree_dispute_reason = fields.Char(
         "Braintree Dispute Reason", readonly=True,
@@ -247,11 +209,25 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
     def __setup__(cls):
         super().__setup__()
         cls.amount.states['readonly'] &= ~Eval('braintree_settlement_needed')
-        cls.amount.depends.append('braintree_settlement_needed')
+
+        cls.braintree_customer.states['readonly'] = (
+                ~Eval('state').in_(['draft', 'submitted', 'approved'])
+                | Eval('braintree_nonce'))
+
+        cls.braintree_customer_method.states['invisible'] |= (
+            Eval('braintree_nonce'))
+        cls.braintree_customer_method.states['readonly'] = (
+            ~Eval('state').in_(['draft', 'submitted', 'approved']))
+
+        cls.braintree_customer_method_selection.states['invisible'] |= (
+            Eval('braintree_nonce'))
+        cls.braintree_customer_method_selection.states['readonly'] = (
+            ~Eval('state').in_(['draft', 'submitted', 'approved']))
+
         cls._buttons.update({
                 'braintree_checkout': {
                     'invisible': (~Eval('state', 'draft').in_(
-                            ['approved', 'processing'])
+                            ['submitted', 'approved', 'processing'])
                         | (Eval('process_method') != 'braintree')
                         | Eval('braintree_nonce')
                         | Eval('braintree_customer_method')),
@@ -273,40 +249,6 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
                     },
                 })
 
-    @fields.depends('journal')
-    def on_change_with_braintree_account(self, name=None):
-        if self.journal and self.journal.process_method == 'braintree':
-            return self.journal.braintree_account.id
-
-    def on_change_party(self):
-        super().on_change_party()
-        self.braintree_customer = None
-
-    @fields.depends('braintree_customer', 'braintree_customer_method')
-    def get_braintree_customer_methods(self):
-        methods = [('', '')]
-        if self.braintree_customer:
-            methods.extend(self.braintree_customer.payment_methods())
-        if (self.braintree_customer_method
-                and self.braintree_customer_method not in dict(methods)):
-            methods.append((
-                    self.braintree_customer_method,
-                    self.braintree_customer_method))
-        return methods
-
-    @fields.depends(
-        'braintree_customer_method_selection', 'braintree_customer_method')
-    def on_change_braintree_customer_method_selection(self):
-        self.braintree_customer_method = (
-            self.braintree_customer_method_selection)
-
-    def get_braintree_customer_method(self, name):
-        return self.braintree_customer_method
-
-    @classmethod
-    def set_braintree_customer_method(cls, payments, name, value):
-        pass
-
     @classmethod
     def default_braintree_settle_payment(cls):
         return True
@@ -330,18 +272,21 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
             ]
 
     @classmethod
-    def validate(cls, payments):
-        super(Payment, cls).validate(payments)
-        for payment in payments:
-            payment.check_braintree_journal()
+    def validate_fields(cls, payments, field_names):
+        super().validate_fields(payments, field_names)
+        cls.check_braintree_journal(payments, field_names)
 
-    def check_braintree_journal(self):
-        if (self.kind != 'receivable'
-                and self.journal.process_method == 'braintree'):
-            raise PaymentValidationError(
-                gettext('account_payment_braintree.msg_braintree_receivable',
-                    journal=self.journal.rec_name,
-                    payment=self.rec_name))
+    @classmethod
+    def check_braintree_journal(cls, payments, field_names=None):
+        if field_names and not (field_names & {'kind', 'journal'}):
+            return
+        for payment in payments:
+            if (payment.kind != 'receivable'
+                    and payment.journal.process_method == 'braintree'):
+                raise PaymentValidationError(gettext(
+                        'account_payment_braintree.msg_braintree_receivable',
+                        journal=payment.journal.rec_name,
+                        payment=payment.rec_name))
 
     @classmethod
     def copy(cls, payments, default=None):
@@ -590,25 +535,30 @@ class PaymentBraintreeRefund(Workflow, ModelSQL, ModelView):
             ],
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
-    amount = fields.Numeric(
-        "Amount", required=True, digits=(16, Eval('currency_digits', 2)),
+            })
+    amount = Monetary(
+        "Amount", currency='currency', digits='currency', required=True,
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['currency_digits', 'state'])
+            })
+    submitted_by = employee_field(
+        "Submitted by",
+        states=['submitted', 'approved', 'processing', 'succeeded', 'failed'])
+    approved_by = employee_field(
+        "Approved by",
+        states=['approved', 'processing', 'succeeded', 'failed'])
 
     state = fields.Selection([
             ('draft', "Draft"),
+            ('submitted', "Submitted"),
             ('approved', "Approved"),
             ('processing', "Processing"),
             ('succeeded', "Succeeded"),
             ('failed', "Failed"),
-            ], "State", readonly=True)
+            ], "State", readonly=True, sort=False)
 
     braintree_transaction_id = fields.Char(
-        "Braintree Transaction ID", readonly=True)
+        "Braintree Transaction ID", readonly=True, strip=False)
 
     braintree_error_message = fields.Text(
         "Braintree Error Message", readonly=True,
@@ -616,15 +566,27 @@ class PaymentBraintreeRefund(Workflow, ModelSQL, ModelView):
             'invisible': ~Eval('braintree_error_message'),
             })
 
-    currency_digits = fields.Function(
-        fields.Integer("Currency Digits"), 'on_change_with_currency_digits')
+    currency = fields.Function(fields.Many2One(
+            'currency.currency', "Currency"),
+        'on_change_with_currency')
+    company = fields.Function(
+        fields.Many2One('company.company', "Company"),
+        'on_change_with_company')
 
     @classmethod
     def __setup__(cls):
         super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t,
+                (t.state, Index.Equality()),
+                where=t.state.in_([
+                        'draft', 'submitted', 'approved', 'processing'])))
         cls.__access__.add('payment')
         cls._transitions |= set((
-                ('draft', 'approved'),
+                ('draft', 'submitted'),
+                ('submitted', 'approved'),
                 ('approved', 'processing'),
                 ('processing', 'succeeded'),
                 ('processing', 'failed'),
@@ -632,25 +594,33 @@ class PaymentBraintreeRefund(Workflow, ModelSQL, ModelView):
                 ))
         cls._buttons.update({
                 'draft': {
-                    'invisible': Eval('state') != 'approved',
+                    'invisible': ~Eval('state').in_(['approved', 'submitted']),
                     'icon': 'tryton-back',
                     'depends': ['state'],
                     },
-                'approve': {
+                'submit': {
                     'invisible': Eval('state') != 'draft',
                     'icon': 'tryton-forward',
                     'depends': ['state'],
                     },
+                'approve': {
+                    'invisible': Eval('state') != 'submitted',
+                    'icon': 'tryton-forward',
+                    'depends': ['state'],
+                    },
                 })
 
     @classmethod
     def default_state(cls):
         return 'draft'
 
-    @fields.depends('payment', '_parent_payment.currency_digits')
-    def on_change_with_currency_digits(self, name=None):
-        if self.payment:
-            return self.payment.currency_digits
+    @fields.depends('payment', '_parent_payment.currency')
+    def on_change_with_currency(self, name=None):
+        return self.payment.currency if self.payment else None
+
+    @fields.depends('payment', '_parent_payment.company')
+    def on_change_with_company(self, name=None):
+        return self.payment.company if self.payment else None
 
     @classmethod
     def copy(cls, refunds, default=None):
@@ -665,12 +635,21 @@ class PaymentBraintreeRefund(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('draft')
+    @reset_employee('submitted_by', 'approved_by')
     def draft(cls, refunds):
         pass
 
+    @classmethod
+    @ModelView.button
+    @Workflow.transition('submitted')
+    @set_employee('submitted_by')
+    def submit(cls, refunds):
+        pass
+
     @classmethod
     @ModelView.button
     @Workflow.transition('approved')
+    @set_employee('approved_by')
     def approve(cls, refunds):
         pass
 
@@ -798,10 +777,11 @@ class PaymentBraintreeAccount(ModelSQL, ModelView):
             ('sandbox', "Sandbox"),
             ('production', "Production"),
             ], "Environment", required=True)
-    merchant_id = fields.Char("Merchant ID", required=True)
-    public_key = fields.Char("Public Key", required=True)
-    private_key = fields.Char("Private Key", required=True)
-    webhook_identifier = fields.Char("Webhook Identifier", readonly=True)
+    merchant_id = fields.Char("Merchant ID", required=True, strip=False)
+    public_key = fields.Char("Public Key", required=True, strip=False)
+    private_key = fields.Char("Private Key", required=True, strip=False)
+    webhook_identifier = fields.Char(
+        "Webhook Identifier", readonly=True, strip=False)
     webhook_endpoint = fields.Function(
         fields.Char(
             "Webhook Endpoint",
@@ -811,8 +791,7 @@ class PaymentBraintreeAccount(ModelSQL, ModelView):
         "Automatic Settlement",
         states={
             'invisible': Eval('environment') != 'sandbox',
-            },
-        depends=['environment'])
+            })
 
     @classmethod
     def __setup__(cls):
@@ -930,21 +909,19 @@ class PaymentBraintreeCustomer(
     __name__ = 'account.payment.braintree.customer'
     _history = True
     party = fields.Many2One(
-        'party.party', "Party", required=True, select=True,
+        'party.party', "Party", required=True,
         states={
             'readonly': (
                 Eval('braintree_customer_id') | Eval('braintree_nonce')),
-            },
-        depends=['braintree_customer_id', 'braintree_nonce'])
+            })
     braintree_account = fields.Many2One(
         'account.payment.braintree.account', "Account", required=True,
         states={
             'readonly': (
                 Eval('braintree_customer_id') | Eval('braintree_nonce')),
-            },
-        depends=['braintree_customer_id', 'braintree_nonce'])
+            })
     braintree_customer_id = fields.Char(
-        "Braintree Customer ID",
+        "Braintree Customer ID", strip=False,
         states={
             'readonly': Eval('braintree_customer_id') & (Eval('id', -1) >= 0),
             })
@@ -954,6 +931,13 @@ class PaymentBraintreeCustomer(
             'invisible': ~Eval('braintree_error_message'),
             })
 
+    identical_customers = fields.Many2Many(
+        'account.payment.braintree.customer.identical',
+        'source', 'target', "Identical Customers", readonly=True,
+        states={
+            'invisible': ~Eval('identical_customers'),
+            })
+
     _payment_methods_cache = Cache(
         'account_payment_braintree_customer.payment_methods',
         duration=config.getint(
@@ -1184,6 +1168,34 @@ class PaymentBraintreeCustomer(
         self._payment_methods_cache.clear()
 
 
+class PaymentBraintreeCustomerIdentical(ModelSQL):
+    "Braintree Customer Identical"
+    __name__ = 'account.payment.braintree.customer.identical'
+    source = fields.Many2One('account.payment.braintree.customer', "Source")
+    target = fields.Many2One('account.payment.braintree.customer', "Target")
+
+    @classmethod
+    def table_query(cls):
+        pool = Pool()
+        Customer = pool.get('account.payment.braintree.customer')
+        source = Customer.__table__()
+        target = Customer.__table__()
+        return (
+            source
+            .join(target, condition=(
+                    source.braintree_customer_id
+                    == target.braintree_customer_id))
+            .select(
+                Literal(0).as_('create_uid'),
+                source.create_date.as_('create_date'),
+                Literal(None).as_('write_uid'),
+                Literal(None).as_('write_date'),
+                sql_pairing(source.id, target.id).as_('id'),
+                source.id.as_('source'),
+                target.id.as_('target'),
+                where=source.id != target.id))
+
+
 class PaymentBraintreeCustomerPaymentMethodDelete(Wizard):
     "Delete Customer Payment Method"
     __name__ = 'account.payment.braintree.customer.payment_method.delete'
